use std::{
    env::{self, VarError},
    error::Error,
    fmt::Debug,
    path::PathBuf,
    str::FromStr,
    time::Duration,
};

use anyhow::{anyhow, Context};

pub struct CaptureConfig {
    pub images: bool,
    pub audio: Option<f64>,
    pub subtitle_language: Option<String>,
    pub audio_language: Option<String>,
}

pub struct CohostIdentityConfig {
    pub email: String,
    pub password: String,
    pub page: String,
}

pub struct PostMetadataConfig {
    pub global_tags: Vec<String>,
    pub draft: bool,
    pub cw: bool,
    pub eighteen_plus: bool,
}

pub struct Config {
    pub capture: CaptureConfig,
    pub cohost: CohostIdentityConfig,
    pub metadata: PostMetadataConfig,
    pub shows_file: PathBuf,
    pub post_interval: Duration,
    pub retry_delay: Duration,
    pub max_retries: Option<u8>,
}

const VAR_PREFIX: &str = "SCREENCAP_BOT_";

fn get_var(name: &str) -> anyhow::Result<Option<String>> {
    let var_name = VAR_PREFIX.to_string() + name;
    match env::var(&var_name) {
        Ok(v) => Ok(Some(v)),
        Err(VarError::NotPresent) => Ok(None),
        Err(e) => Err(e).with_context(|| format!("Failed to read variable {}", var_name)),
    }
}

fn parse_var<E: Debug + Send + Sync + Error + 'static, T: FromStr<Err = E>>(
    name: &str,
) -> anyhow::Result<Option<T>> {
    get_var(name)?
        .map(|s| {
            s.parse()
                .with_context(|| format!("Failed to parse variable {}{}", VAR_PREFIX, name))
        })
        .transpose()
}

fn require_var(name: &str) -> anyhow::Result<String> {
    get_var(name)?.ok_or_else(|| anyhow!("{}{} must be set", VAR_PREFIX, name))
}

fn get_language_code_var<F: FnOnce() -> Option<String>>(
    name: &str,
    default: F,
) -> anyhow::Result<Option<String>> {
    match get_var(name)? {
        Some(s) => {
            if s.is_ascii() && s.len() == 3 {
                Ok(Some(s))
            } else if s.is_empty() {
                Ok(None)
            } else {
                Err(anyhow!(
                    "{}{} must be an ISO-639-2 three-letter language code",
                    VAR_PREFIX,
                    name
                ))
            }
        }
        None => Ok(default()),
    }
}

pub fn load() -> anyhow::Result<Config> {
    let capture_images = parse_var("CAPTURE_IMAGES")?.unwrap_or(true);

    let capture_audio_duration = match parse_var::<_, f64>("CAPTURE_AUDIO_DURATION")? {
        Some(d) => {
            if !d.is_finite() {
                return Err(anyhow!(
                    "non-finite float value for {}CAPTURE_AUDIO_DURATION",
                    VAR_PREFIX
                ));
            } else if d >= 1.0 {
                Some(d)
            } else if d > 0.0 {
                return Err(anyhow!(
                    "cannot capture audio clips less than 1 second long"
                ));
            } else if d == 0.0 {
                None
            } else {
                return Err(anyhow!(
                    "{}CAPTURE_AUDIO_DURATION cannot be negative",
                    VAR_PREFIX
                ));
            }
        }
        None => None,
    };

    if let (false, None) = (capture_images, capture_audio_duration) {
        return Err(anyhow!(
            "At least one of image capture and audio capture must be enabled!"
        ));
    }

    Ok(Config {
        capture: CaptureConfig {
            images: capture_images,
            audio: capture_audio_duration,
            subtitle_language: get_language_code_var("SUBTITLE_LANGUAGE", || {
                Some(String::from("eng"))
            })?,
            audio_language: get_language_code_var("AUDIO_LANGUAGE", || None)?,
        },
        cohost: CohostIdentityConfig {
            email: require_var("COHOST_EMAIL")?,
            password: require_var("COHOST_PASSWORD")?,
            page: require_var("COHOST_PAGE")?,
        },
        metadata: PostMetadataConfig {
            global_tags: get_var("GLOBAL_TAGS")?
                .map(|s| s.split(',').map(String::from).collect())
                .unwrap_or_default(),
            draft: parse_var("COHOST_DRAFT")?.unwrap_or(false),
            cw: parse_var("COHOST_CW")?.unwrap_or(capture_images),
            eighteen_plus: parse_var("18PLUS")?.unwrap_or(false),
        },
        shows_file: parse_var("SHOWS_FILE")?.unwrap_or_else(|| PathBuf::from("./shows.yaml")),
        post_interval: Duration::from_secs(parse_var("POST_INTERVAL")?.unwrap_or_default()),
        retry_delay: Duration::from_secs(parse_var("RETRY_DELAY")?.unwrap_or(30)),
        max_retries: parse_var::<_, i8>("MAX_RETRIES")?
            .unwrap_or(5)
            .try_into()
            .ok(),
    })
}
